﻿using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// 윈도우즈 헬퍼
    /// </summary>
    public class WindowsHelper
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Structure
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 키보드 저수준 후킹 구조체 - KBDLLHOOKSTRUCT

        /// <summary>
        /// 키보드 저수준 후킹 구조체
        /// </summary>
        private struct KBDLLHOOKSTRUCT
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Field
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region Field

            /// <summary>
            /// 가상 키 코드
            /// </summary>
            public int VirtualKeyCode;

            /// <summary>
            /// 스캔 코드
            /// </summary>
            public int ScanCode;

            /// <summary>
            /// 플래그
            /// </summary>
            public int Flags;

            /// <summary>
            /// 시간
            /// </summary>
            public int Time;

            /// <summary>
            /// 부가 정보
            /// </summary>
            public int ExtraInfo;

            #endregion
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Delegate
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 키보드 후킹 처리 대리자 - ProcessKeyboardHookDelegate(code, wordParameter, longParameter)

        /// <summary>
        /// 키보드 후킹 처리 대리자
        /// </summary>
        /// <param name="code">코드</param>
        /// <param name="wordParameter">WORD 매개 변수</param>
        /// <param name="longParameter">LONG 매개 변수</param>
        /// <returns>처리 결과</returns>
        private delegate int ProcessKeyboardHookDelegate(int code, int wordParameter, ref KBDLLHOOKSTRUCT longParameter);

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Import
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 윈도우 후킹 설정하기 - SetWindowsHookEx(hookID, processKeyboardHookDelegate, moduleHandle, threadID)

        /// <summary>
        /// 윈도우 후킹 설정하기
        /// </summary>
        /// <param name="hookID">후킹 ID</param>
        /// <param name="processKeyboardHookDelegate">키보드 후킹 처리 대리자</param>
        /// <param name="moduleHandle">모듈 핸들</param>
        /// <param name="threadID">스레드 ID</param>
        /// <returns>처리 결과</returns>
        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern int SetWindowsHookEx
        (
            int                         hookID,
            ProcessKeyboardHookDelegate processKeyboardHookDelegate,
            IntPtr                      moduleHandle,
            uint                        threadID
        );

        #endregion
        #region 윈도우 후킹 해제하기 - UnhookWindowsHookEx(hookHandle)

        /// <summary>
        /// 윈도우 후킹 해제하기
        /// </summary>
        /// <param name="hookHandle">후킹 핸들</param>
        /// <returns>처리 결과</returns>
        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool UnhookWindowsHookEx(int hookHandle);

        #endregion
        #region 다음 후킹 호출하기 - CallNextHookEx(hookHandle, code, wordParameter, longParameter)

        /// <summary>
        /// 다음 후킹 호출하기
        /// </summary>
        /// <param name="hookHandle">후킹 핸들</param>
        /// <param name="code">코드</param>
        /// <param name="wordParameter">WORD 매개 변수</param>
        /// <param name="longParameter">LONG 매개 변수</param>
        /// <returns>처리 결과</returns>
        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern int CallNextHookEx(int hookHandle, int code, int wordParameter, ref KBDLLHOOKSTRUCT longParameter);

        #endregion
        #region 모듈 핸들 구하기 - GetModuleHandle(modulName)

        /// <summary>
        /// 모듈 핸들 구하기
        /// </summary>
        /// <param name="modulName">모듈명</param>
        /// <returns>모듈 핸들</returns>
        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern IntPtr GetModuleHandle(string modulName);

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 키보드 후킹 처리 대리자
        /// </summary>
        private static ProcessKeyboardHookDelegate _processKeyboardHookDelegate = ProcessKeyboardHook;

        /// <summary>
        /// 후킹 ID
        /// </summary>
        private static int _hookID = 0;

        /// <summary>
        /// 태스크 관리자 이용 가능 여부
        /// </summary>
        private static bool _taskManagerEnabled = true;

        /// <summary>
        /// 특수 키보드 입력 이용 가능 여부
        /// </summary>
        private static bool _specialKeyboardInputEnabled = true;

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// WH_KEYBOARD_LL
        /// </summary>
        private const int WH_KEYBOARD_LL = 13;

        /// <summary>
        /// WM_KEYDOWN
        /// </summary>
        private const int WM_KEYDOWN = 0x0100;

        /// <summary>
        /// WM_KEYUP
        /// </summary>
        private const int WM_KEYUP = 0x0101;

        /// <summary>
        /// WM_SYSKEYDOWN
        /// </summary>
        private const int WM_SYSKEYDOWN = 0x0104;

        /// <summary>
        /// WM_SYSKEYUP
        /// </summary>
        private const int WM_SYSKEYUP = 0x0105;


        /// <summary>
        /// 부모 종료 버튼
        /// </summary>
        private Button parentFinishButton;

        /// <summary>
        /// 잠금 여부
        /// </summary>
        private bool isLocked = false;

        /// <summary>
        /// 화면 리스트
        /// </summary>
        private List<Screen> screenList = null;

        /// <summary>
        /// 폼 리스트
        /// </summary>
        private List<WallForm> formList = null;

        #endregion
        
        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - WindowsHelper(parentFinishButton)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="parentFinishButton">부모 종료 버튼</param>
        public WindowsHelper(Button parentFinishButton)
        {
            this.parentFinishButton = parentFinishButton;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 태스크 관리자 비활성화 하기 - DisableTaskManager()

        /// <summary>
        /// 태스크 관리자 비활성화 하기
        /// </summary>
        /// <remarks>Ctrl + Alt + Delete 키 입력을 거부한다.</remarks>
        public static void DisableTaskManager()
        {
            if(!_taskManagerEnabled)
            {
                return;
            }

            RegistryKey registryKey;
            int         keyValue = 1;
            string      subKey   = "Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\System";

            registryKey = Registry.CurrentUser.CreateSubKey(subKey);

            registryKey.SetValue("DisableTaskMgr", keyValue);

            registryKey.Close();

            _taskManagerEnabled = false;
        }

        #endregion
        #region 태스크 관리자 활성화 하기 - EnableTaskManager()

        /// <summary>
        /// 태스크 관리자 활성화 하기
        /// </summary>
        /// <remarks>Ctrl + Alt + Delete 키 입력을 허용한다.</remarks>
        public static void EnableTaskManager()
        {
            if(_taskManagerEnabled)
            {
                return;
            }

            RegistryKey currentUserRegistryKey = Registry.CurrentUser;
            string      subKey                 = "Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\System";
            RegistryKey registryKey            = currentUserRegistryKey.OpenSubKey(subKey);

            if(registryKey != null)
            {
                currentUserRegistryKey.DeleteSubKeyTree(subKey);
            }

            _taskManagerEnabled = true;
        }

        #endregion

        #region 특수 키보드 입력 비활성 화기 - DisableSpecialKeyboardInput()

        /// <summary>
        /// 특수 키보드 입력 비활성 화기
        /// </summary>
        public static void DisableSpecialKeyboardInput()
        {
            if(!_specialKeyboardInputEnabled)
            {
                return;
            }

            _hookID = SetHook(_processKeyboardHookDelegate);

            _specialKeyboardInputEnabled = false;
        }

        #endregion
        #region 특수 키보드 입력 활성화 하기 - EnableSpecialKeyboardInput()

        /// <summary>
        /// 특수 키보드 입력 활성화 하기
        /// </summary>
        public static void EnableSpecialKeyboardInput()
        {
            if(_specialKeyboardInputEnabled)
            {
                return;
            }

            UnhookWindowsHookEx(_hookID);

            _specialKeyboardInputEnabled = true;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region 후킹 설정하기 - SetHook(processKeyboardHookDelegate)

        /// <summary>
        /// 후킹 설정하기
        /// </summary>
        /// <param name="processKeyboardHookDelegate">키보드 후킹 처리 대리자</param>
        /// <returns>처리 결과</returns>
        private static int SetHook(ProcessKeyboardHookDelegate processKeyboardHookDelegate)
        {
            using(Process process = Process.GetCurrentProcess())
            {
                using(ProcessModule processModule = process.MainModule)
                {
                    return SetWindowsHookEx(WH_KEYBOARD_LL, processKeyboardHookDelegate, GetModuleHandle(processModule.ModuleName), 0);
                }
            }
        }

        #endregion
        #region 키보드 후킹 처리하기 - ProcessKeyboardHook(code, wordParameter, longParameter)

        /// <summary>
        /// 키보드 후킹 처리하기
        /// </summary>
        /// <param name="code">코드</param>
        /// <param name="wordParameter">WORD 매개 변수</param>
        /// <param name="longParameter">LONG 매개 변수</param>
        /// <returns>처리 결과</returns>
        private static int ProcessKeyboardHook(int code, int wordParameter, ref KBDLLHOOKSTRUCT longParameter)
        {
            bool result = false;

            switch(wordParameter)
            {
                case WM_KEYDOWN    :
                case WM_KEYUP      :
                case WM_SYSKEYDOWN :
                case WM_SYSKEYUP   :

                    result = ((longParameter.VirtualKeyCode == 0x09) && (longParameter.Flags == 0x20)) || // Alt + Tab
                             ((longParameter.VirtualKeyCode == 0x1B) && (longParameter.Flags == 0x20)) || // Alt + Esc
                             ((longParameter.VirtualKeyCode == 0x1B) && (longParameter.Flags == 0x00)) || // Ctrl + Esc
                             ((longParameter.VirtualKeyCode == 0x5B) && (longParameter.Flags == 0x01)) || // Left Windows Key
                             ((longParameter.VirtualKeyCode == 0x5C) && (longParameter.Flags == 0x01)) || // Right Windows Key
                             ((longParameter.VirtualKeyCode == 0x73) && (longParameter.Flags == 0x20));   // Alt + F4

                    break;
            }

            if(result == true)
            {
                return 1;
            }
            else
            {
                return CallNextHookEx(0, code, wordParameter, ref longParameter);
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 화면 잠그기 - LockScreen()

        /// <summary>
        /// 화면 잠그기
        /// </summary>
        public void LockScreen()
        {
            if(this.isLocked)
            {
                return;
            }

            DisableTaskManager();

            DisableSpecialKeyboardInput();

            this.screenList = Screen.AllScreens.ToList();

            this.formList = new List<WallForm>();

            for(int i = 0; i < this.screenList.Count; i++)
            {
                WallForm form = new WallForm();

                form.ParentFinishButton = this.parentFinishButton;

                form.FormBorderStyle = FormBorderStyle.None;
                form.WindowState     = FormWindowState.Maximized;
                form.StartPosition   = FormStartPosition.Manual;
                form.Left            = this.screenList[i].WorkingArea.Left;
                form.Top             = this.screenList[i].WorkingArea.Top;
                form.Width           = this.screenList[i].WorkingArea.Width;
                form.Height          = this.screenList[i].WorkingArea.Height;
                form.TopMost         = true;

                this.formList.Add(form);

                form.Show();
            }

            this.isLocked = true;
        }

        #endregion
        #region 화면 잠금 취소하기 - UnlockScreen()

        /// <summary>
        /// 화면 잠금 취소하기
        /// </summary>
        public void UnlockScreen()
        {
            if(!this.isLocked)
            {
                return;
            }

            EnableTaskManager();

            EnableSpecialKeyboardInput();

            for(int i = this.formList.Count - 1; i > -1; i--)
            {
                WallForm form = this.formList[i];

                this.formList.Remove(form);

                form.Hide();

                form.Dispose();
            }

            this.isLocked = false;
        }

        #endregion
    }
}